(* : Title: SeriesSolve *)

(* : Context: Applications`SeriesSolve` *)

(* : Author: Jean C. Piquette *)

(* : Mathematica version: 3.0 *)

BeginPackage["Applications`SeriesSolve`"]

SeriesSolve::usage = 
"SeriesSolve[equation, variable, small parameter, (options)] finds one
root (or one root set) of equation (may be a list of equations), for the
variable (may be a list of variables), expressed in terms of a Maclaurin
series expansion in the small parameter, for those roots that possess
a Maclaurin series. The option Order specifies the desired maximum order
of the expansion(s) (default is 2). The option Root specifies which of
the sequentially numbered roots of the 'zeroth-order' equation is used
to form the zeorth term of each Maclaurin series. (The zeroth order
equation is the equation that results from setting the small parameter
to zero in equation.) Default value for Root is 1. The option SpeedSimp
is used for selecting maximum speed (if True) or maximum simplification
of results (if False). Default value of SpeedSimp is True."

    
     

(* NOTE: SET debug = 1 BELOW TO OBTAIN DEBUGGING PRINTOUTS *)


Options[SeriesSolve] = {Order->2,Root->1,SpeedSimp->True}

Begin["`Private`"]

(* ERROR MESSAGES *)

SeriesSolve::speedsimpwrong="speedsimp variable must be set either to True (for speed) or False (for greatest simplification)"
SeriesSolve::orderwrong="desired order must be a positive integer"
SeriesSolve::ordernonnegative="desired order must be a positive integer"
SeriesSolve::multiplerootwrong="desired multiple root number must be a positive integer"
SeriesSolve::multiplerootzero="multiple root number cannot be 0"
SeriesSolve::multiplerootnonnegative="desired multiple root number must be a positive integer"
SeriesSolve::notwellformed="equations are not well-formed"
SeriesSolve::nosmallparam="equations do not contain specified small parameter `1`"
SeriesSolve::novars="variables of interest are not present in zeroth-order equations...no maclaurin solutions"
SeriesSolve::unequal="number of variables must equal number of equations"
SeriesSolve::insufficientroots="chosen multiple root exceeds number of available roots = `1`"
SeriesSolve::nomacsols="no maclaurin solutions"
SeriesSolve::nomacsoldelta="no maclaurin solutions to delta = `1` numerical accuracy (but solutions may be found to lower order); also, try exact numerical values or purely symbolic computation"
SeriesSolve::backsubfail="Solution obtained for zeroth-order equations does not satisfy them on backsubstitution; if multiple roots are available, try a different root"

SeriesSolve[testeq_,dD_,tT3_,opts___] :=
    Module[{c,d,npart,test,j,expansion,ans,maclaurt,maclaurin,maclaurhold,
        accum,pos,part,lenpart,newpart,knew,numbeqs,numbvars,expansionhold,
        k,kexp,kmin,m,n,n2,m2,temp,testeq2,dD2,exptest,expmax,debug,soltest,
        testlist,x,y,simplify,loop,delta,tsub,precisionflag,accuracy,
        testflag,testflag2},

{jmax,nroot,speedsimp}={Order,Root,SpeedSimp}/.{opts}/.Options[SeriesSolve];
Off[Solve::ifun];
Off[Solve::tdep];
If[(speedsimp==True)||(speedsimp==False),
    ,
    Message[SeriesSolve::speedsimpwrong];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
    Message[SeriesSolve::speedsimpwrong];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];

simplify[x_]:=(y=Together[x];Return[Expand[Numerator[y]]/Denominator[y]]);
    If[speedsimp==False,simplify[x_]:=Return[Simplify[x]]
      ];

debug=0;loop=0;accuracy=20;delta=10^-accuracy; 
    tsub=tT3; (* SET delta FOR DESIRED REAL-NUMBER ZERO CHECKING *)

If[testeq[[0]] != List,
    temp={};temp = Append[temp,testeq];testeq2=temp;,
    testeq2=testeq,temp={};temp = Append[temp,testeq];testeq2=temp
  ];

If[dD[[0]] != List,
    temp={};temp = Append[temp,dD];dD2=temp,
    dD2=dD,
    temp={};temp = Append[temp,dD];dD2=temp
  ];


If[IntegerQ[jmax],
    ,
    Message[SeriesSolve::orderwrong];
	On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
    Message[SeriesSolve::orderwrong];
	On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];


If[IntegerQ[nroot],
    ,
    Message[SeriesSolve::multiplerootwrong];
        On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
    Message[SeriesSolve::multiplerootwrong];
        On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];
            
If[nroot==0,
    Message[SeriesSolve::multiplerootzero];
	On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];
If[jmax<0,
    Message[SeriesSolve::ordernonnegative];
	On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];
            
        

If[nroot<0,
    Message[SeriesSolve::multiplerootnonnegative];
        On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];

numbeqs=Length[testeq2];numbvars=Length[dD2];




If[testeq2[[0]]==List,
    ,
    If[testeq2[[0]]!=Equal,
        Message[SeriesSolve::notwellformed];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
        ,
        Message[SeriesSolve::notwellformed];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
      ];testeq=testeq2[[1]]-testeq2[[2]],      
    If[testeq2[[0]]!=Equal,
        Message[SeriesSolve::notwellformed];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
        ,
        Message[SeriesSolve::notwellformed];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
      ];testeq=testeq2[[1]]-testeq2[[2]]          
   ];
For[n=1,n<numbeqs+1,n++,
    If[testeq2[[n,0]]!=Equal,
        Message[SeriesSolve::notwellformed];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
        ,
        Message[SeriesSolve::notwellformed];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
      ]
    ];

For[n=1,n<numbeqs+1,
    n++,testeq2[[n]]=testeq2[[n,1]]-testeq2[[n,2]]];
 





precisionflag=Infinity;
For[n=1,n<numbeqs+1,n++,
    If[Precision[testeq2[[n]]]!=Infinity,
        precisionflag=0
      ]
   ];



testflag=0;
For[n=1,n<numbeqs+1,n++,
    If[!FreeQ[testeq2[[n]],
          tT3],
        testflag=1
      ];
    ];
If[testflag==0,
    Message[SeriesSolve::nosmallparam,tT3];
        On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];



(*                
For[n=1,n<numbeqs+1,n++,
    testflag=0;
    For[m=1,m<numbvars+1,m++,
        If[!FreeQ[(testeq2[[n]]/.tT3->0),
            dD2[[m]]],
            testflag=1
          ];
	];
     If[testflag==0,
         Message[SeriesSolve::novars];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
        ]
    ];
*)	

For[n=1,n<numbvars+1,n++,
    testflag2[n]=0;
    For[m=1,m<numbeqs+1,m++,
        If[!FreeQ[(testeq2[[m]]/.tT3->0),
            dD2[[n]]],
            testflag2[n]=1
          ]
        ];
    If[testflag2[n]==0,
        Message[SeriesSolve::novars];
		On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
      ]
    ];








If[debug==1,Print[numbvars]];

If[numbvars != numbeqs,
    Message[SeriesSolve::unequal];
        On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
  ];
            
expmax=0;

For[
    For[n=1,n<numbvars+1,n++,maclaurt[n]=0
       ];
    j=0, j<jmax+1, j++,
    For[n=1,n<numbvars+1,n++,
        maclaurt[n]=maclaurt[n]+c[j,n] (tT3)^j
       ]; 
    If[debug==1, 
        For[n=1,n<numbvars+1,n++,
            Print["function series:  " ,dD2[[n]]," ->  " ,maclaurt[n]
                 ]
           ]
      ];
     For[n=1,n<numbeqs+1,n++,
         test[n]=testeq2[[n]]
        ];
     For[n=1,n<numbeqs+1,n++,
         For[np=1,np<numbvars+1,np++,
             test[n] = test[n]/.dD2[[np]]->maclaurt[np]]
            ];
         If[debug==1, 
             For[n=1,n<numbeqs+1,n++,
                 Print[test[n]
                      ]
                ]
           ];









            k = 0 ;                        
            For[n=1,n<numbeqs+1,n++,kexp[n]=k
               ];
            For[n=1,n<numbeqs+1,n++,
                kmin[n]=Exponent[test[n],tT3,Min];
                If[kexp[n]<kmin[n],
                kexp[n]=kmin[n]
                  ]
               ];
            If[debug==1,
                For[n=1,n<numbeqs+1,n++,
                    Print["k,n,kmin,test[n]= ",k,n,kmin[n],
                          test[n]
                         ]
                   ]
              ];



            For[n=1,n<numbeqs+1,n++,
               While[True,
                   If[debug==1,
                       Print["examining equation number: ",n
                            ]
                     ];
                    expansion[n] = 
                        PowerExpand[Normal[
                               Series[test[n],{tT3,0,kexp[n]}
                                     ]
                                          ]
                                   ];
	              If[debug==1,
                        Print["intermediate expansion[n]: ",
                            expansion[n]
                             ]
                      ];
                    testflag=0;
                    For[m=1,m<numbeqs+1,m++,
                        If[debug==1,
                            Print["n,j,m,expansion[n],c[j,m]",
                                n,j,m,expansion[n],c[j,m]
                                 ]
                           ];
	                   If[FreeQ[expansion[n],c[j,m]],,testflag=1;Break[]];
			     	];
			  If[testflag==0,
                            kexp[n]=kexp[n]+1;
                            If[debug==1,
                                Print["increasing exponent 
                                       because equation has no
                                       c[j,m]"
                                     ]
                              ];Continue[]
					 
			     ];
					

                   expansionhold[n]=expansion[n];
                   For[m=1, m<j+1, m++, 
                       d[n,m]=simplify[PowerExpand[Coefficient[
                           expansion[n],(tT3)^m
                                                              ]
                                                  ]
                                      ]
                      ];
                   For[expansion[n]=0; m=1, m<j+1, m++, 
                       expansion[n] = expansion[n] 
                           + d[n,m] tT3^m
                      ];
                   d[n,0]=simplify[PowerExpand[Expand[
                       expansionhold[n]-expansion[n]
                                                      ]
                                               ]
                                  ];
                   expansion[n]=d[n,0]+expansion[n];
                   expansion[n]=PowerExpand[expansion[n]];              
		       If[expansion[n]==0,
                       kexp[n]=kexp[n]+1;Continue[]
                     ];
			 Break[]
                 ]; (* END FOR ON n*)
                 ]; (* END WHILE *)



       If[debug==1, 
           For[n=1,n<numbvars+1,n++,
               Print["equation number: ",n];
               Print["expansion exponent: ",kexp[n]," equation series:
                       ",expansion[n]
                     ]
               ]
           ];

                         
        Label[cont1]; If[debug==1, Print["at cont1"]]; 



                         
        eqlist={};
        For[n=1,n<numbeqs+1,n++,
            eqlist=Append[eqlist,expansion[n]==0]
           ];(* eqlist=eqlist/.tT3->1; *)
        If[debug==1, 
            Print["eqlist: ",eqlist
                 ]
          ]; 
                                

        anslist={};
        For[n=1,n<numbvars+1,n++,
            anslist=Append[anslist,c[j,n]]
           ];
        If[debug==1, 
            Print["anslist: ",anslist]
          ];
        For[n=1,n<numbeqs+1,n++,
            exptest=Exponent[maclaurt[n],tT3];
            If[exptest>expmax,
                expmax=exptest
              ]; 
             If[debug==1, 
                 Print["expmax=",expmax," maclaurt= ",maclaurt[n]
                      ]
               ]
            ]; 

         eqlist2={};
         For[n=1,n<numbeqs+1,n++,
             eqlist2=Append[eqlist2,
                        Numerator[simplify[Chop[expansion[n],delta
                                               ]
                                          ]
                                 ]
                           ]
            ];
                         


	ans= Solve[eqlist,anslist];

        If[debug==1, 
            Print[ans]]; 

        If[debug==1, 
            Print["length of ans:",
            Length[ans]
                 ]
          ];


          loop=loop+1;
          If[loop==1,
              If[nroot>Length[ans],
                 Message[SeriesSolve::insufficientroots,Length[ans]]
                     ;On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
                ]
            ];

         npart=1;
         If[Length[ans]>1,
             npart=nroot
           ];

         For[n=1,n<numbeqs+1,n++,
             pos=Position[ans,c[j,n]];
             If[debug==1, 
                 Print["pos:",pos]
               ]; 
             part=pos[[npart]];
             If[debug==1,
                 Print[" part:  ",part]
               ]; 
             lenpart=Length[part];
             If[debug==1,
                 Print[" lenpart: ",lenpart]
               ];

          part=ReplacePart[part,2,lenpart];
          If[debug==1, 
              Print[" part:",part];
              Print[" ans[[part]]= ",Extract[ans,part]
                   ]
            ];  
          For[m=1,m<numbeqs+1,m++,
              If[debug==1, 
                  Print[j,m,"compared:",Extract[ans,part],", "c[j,m]
                       ];
                  Print[FreeQ[Extract[ans,part],c[j,m]
                       ]
                ]
              ];
           If[FreeQ[Extract[ans,part],c[j,m]],
             ,
             If[precisionflag==Infinity,Message[SeriesSolve::nomacsols],
                 Message[SeriesSolve::nomacsoldelta,delta]
               ];On[Solve::ifun];On[Solve::tdep];Return[maclaurin]]]];
             For[n=1,n<numbeqs+1,n++,
                 pos=Position[ans,c[j,n]];
                 If[debug==1, 
                     Print["pos:",pos]
                   ];
                 part=pos[[npart]];
                  If[debug==1,
                      Print[" part:  ",part]
                    ];
                   lenpart=Length[part];
                   If[debug==1, Print[" lenpart: ",lenpart]
                     ]; 
                   part=ReplacePart[part,2,lenpart];
                   If[debug==1, 
                       Print[" part:",part];  
                       Print[" ans[[part]]= ",Extract[ans,part]
                            ]
                     ]; 
                   c[j,n]=Chop[Extract[ans,part],delta];
               ]; 

            If[debug==1, 
                For[n=1,n<numbeqs+1,n++,
                    Print["coefficient solution:  ", "
                          c[",j,n,"] -> ",c[j,n]
                         ]
                   ]
               ];

        
      testlist={True};
      For[m2=1,m2<numbeqs,m2++,
          testlist=Append[testlist,True]
         ];
      If[debug==1,Print[testlist]];




       soltest={};
       For[n=1,n<numbeqs+1,n++,
           If[precisionflag==Infinity,
               soltest=Append[soltest,
                           Chop[PowerExpand[Simplify[PowerExpand[
                               eqlist2[[n]]/.tT3->tsub
                                                                ]
                                                     ]
                                           ]
                          ,delta]==0
                             ],
                soltest=Append[soltest,
                          Chop[PowerExpand[Simplify[PowerExpand[
                              N[eqlist2[[n]]]/.tT3->tsub
                                                               ]
                                                   ]
                                           ]
                           ,delta]==0
                               ]
               ]
            ];



                        
      soltest=soltest/.(0.->0);
      If[debug==1,
          Print["soltest: ",soltest]
         ];
      If[debug==1,
          Print["testing zeroth order"]
        ];




      While[True,
          If[precisionflag!=Infinity,
              Break[]
            ];
          If[soltest!=testlist,
              Message[SeriesSolve::backsubfail];
                   On[Solve::ifun];On[Solve::tdep];Return[maclaurin],
              Break[],
              Message[SeriesSolve::backsubfail];
                   On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
             ]; (* END IF *) 
          Break[];
            ]; (* END WHILE*)


      If[debug==1,
          Print["result of backsubstitution: ",soltest]
        ];


  

      For[n=1,n<numbeqs+1,n++,
          c[j,n]= Simplify[PowerExpand[c[j,n]]];
          If[debug==1,
              Print["simplified coefficient solution: c[j,n] for j=
                    ",j," n=",n," : ",c[j,n]]
                   ]
             ]
          ];

    
       If[debug==1,
           Print["testing c[j,n] solutions for presence of expansion
                 parameter tT3"
                 ]
          ];
        For[j=1,j<jmax+1,j++,
            For[n=1,n<numbeqs+1,n++,
                If[debug==1,
                    Print[c[j,n],tT3,FreeQ[c[j,n],tT3]
                         ]
                  ];
                If[FreeQ[c[j,n],tT3],
                    ,
                    If[precisionflag==Infinity,
                        Message[SeriesSolve::nomacsols],
                        Message[SeriesSolve::nomacsoldelta,delta]
                              
                       ];
                On[Solve::ifun];On[Solve::tdep];Return[maclaurin]
                   ]
              ]
            ];
              

                                                                
      For[n=1,n<numbeqs+1,n++,
          maclaurt[n]=Expand[maclaurt[n]]
         ];
      For[n=1,n<numbeqs+1,n++,
          maclaurt[n]=Collect[maclaurt[n],tT3]
         ];
      For[n=1,n<numbeqs+1,n++,
          maclaurhold[n]=maclaurt[n]
         ];
      For[j=1, j<jmax+1, j++,
          For[n=1,n<numbeqs+1,n++,
              If[speedsimp==True,
                  c[j,n]=PowerExpand[Coefficient[maclaurt[n],(tT3)^j]],
                  c[j,n]=Simplify[PowerExpand[Coefficient[
                     maclaurt[n],(tT3)^j
                                                         ]
                                             ]
                                 ]
                ]
              ]
           ];


     For[
         For[n=1,n<numbeqs+1,n++,
             maclaurt[n]=0
            ];
         j=0, j<jmax+1, j++,
             For[n=1,n<numbeqs+1,n++,
                 maclaurt[n] = maclaurt[n] + c[j,n] tT3^j
                ]
        ];


     For[n=1,n<numbeqs+1,n++,
         If[speedsimp==True,
             c[0,n]=PowerExpand[Expand[maclaurhold[n]-maclaurt[n]]],
             c[0,n]=Simplify[PowerExpand[Expand[
                 maclaurhold[n]-maclaurt[n]
                                               ]
                                         ]
                             ]
            ]
         ]; 


      accum={};

      For[n=1,n<numbeqs+1,n++,
          accum=Append[accum,dD2[[n]]->c[0,n]+maclaurt[n]]
         ];
     maclaurin=Chop[accum,delta];
    For[n=1,n<numbeqs+1,n++,
        maclaurin[[n,2]]=maclaurin[[n,2]]+O[tT3]^(jmax+1)];
On[Solve::ifun];
On[Solve::tdep];
maclaurin]
           
End[]
EndPackage[]



